/* リスト4.8 Flight ShootARのプログラム */

import com.shigeodayo.ardrone.processing.*;  // ARDroneForP5
import jp.nyatla.nyar4psg.*;                 // NyAR4psg

final int MARKER_NUM = 10; // NyIDマーカの数（ID:0～9）
final int MARKER_SIZE = 150;
final int IMAGE_WIDTH = 320; 
final int IMAGE_HEIGHT = 240;

ARDroneForP5 ardrone;  // AR.Drone
MultiMarker  ar;       // ARマーカ
Target[]     targets;                      // ターゲット
PImage       sight, hole, star;            // 画像（サイト，穴，星）
int          centerId, shootDownCount = 0; // 中央にあるマーカのID，撃破数
float        step = 0.0;                   // ロックオンリングの回転量

void setup() {
  size(640, 480, P3D);  // 画面サイズの設定

  ardrone = new ARDroneForP5(); // AR.Droneをコントロールするオブジェクトを生成
  ardrone.connect();            // AR.Droneの本体の制御系に接続
  ardrone.connectNav();         // AR.Droneのセンサ情報系に接続
  ardrone.connectVideo();       // AR.Droneの画像情報系に接続
  ardrone.start();              // 接続を開始

  // ARマーカ認識の準備
  ar = new MultiMarker(this, IMAGE_WIDTH, IMAGE_HEIGHT, "camera_para.dat", NyAR4PsgConfig.CONFIG_PSG);
  ar.setThreshold(40);  // マーカ検出のための閾値（1～255）：環境の明るさに応じて調整
  for (int i = 0; i < MARKER_NUM; i++) 
    ar.addNyIdMarker(i, MARKER_SIZE);  // IDマーカのIDと物理サイズの設定

  // 画像の準備
  sight = loadImage("sight.png");
  hole  = loadImage("hole.png");
  star  = loadImage("star.png");

  // ターゲットの準備
  targets = new Target[MARKER_NUM];  // マーカの数だけターゲットを生成
  for (int i = 0; i < targets.length; i++) 
    targets[i] = new Target(i);
}

void draw() { 
  // カメラ画像の取得と描画
  PImage img = ardrone.getVideoImage(true);
  if (img == null) return;
  imageMode(CORNER);
  ar.drawBackground(img);

  // マーカの認識とターゲットの描画
  ar.detect(img);
  centerId = -1;  
  for (int i = 0; i < targets.length; i++) {
    if (!ar.isExistMarker(i)) continue;
    ar.beginTransform(i);
    targets[i].draw();
    // マーカが画面中央にあり，ターゲットが生きていれば
    if (isCenter(getCenterOfMarker(i)) && targets[i].isAlive()) {
      centerId = i;
      drawLockOnRing(); // ロックオン状態を示すリングを描画
    }
    ar.endTransform();
  }
  
  hint(DISABLE_DEPTH_TEST); // ここから2次元の描画なのでZバッファを無効化

  // サイトの描画
  imageMode(CORNER);
  image(sight, width / 2 - 100, height / 2 - 100);
  
  // バッテリメータの描画
  int battery = ardrone.getBatteryPercentage();
  fill(255, 255, 255, 10);
  rect(10, 10, 200, 20);
  fill(0, 255, 0, 80);
  rect(10, 10, battery * 2, 20);

  // 撃破数に応じて星を描画
  for (int i = 0; i < shootDownCount; i++) 
    image(star, 10, 40 + i * 35);

  // 高度計の描画
  float altitude = ardrone.getAltitude();  // AR.Droneから高度情報を取得
  stroke(0, 255, 0);
  line(600, 40, 600, 440);  // 縦線
  for (int i = 0; i < 9; i++) 
    line(580, 40 + i * 50, 620, 40 + i * 50); // 横線
  fill(0, 255, 0);
  triangle(570, (height - 40) - altitude / 5, 560, (height - 40) - altitude / 5 + 10, 560, (height - 40) - altitude / 5 - 10);
    
  hint(ENABLE_DEPTH_TEST); // 2次元の描画はここまでなので，再度Zバッファを有効化
}

// キーボード入力による操作
void keyPressed() {
  if (key == CODED) {
    if      (keyCode == UP)      ardrone.forward();  // [↑] 前方移動
    else if (keyCode == DOWN)    ardrone.backward(); // [↓] 後方移動
    else if (keyCode == LEFT)    ardrone.goLeft();   // [←] 左移動
    else if (keyCode == RIGHT)   ardrone.goRight();  // [→] 右移動
    else if (keyCode == SHIFT)   ardrone.takeOff();  // [Shift] 離陸
    else if (keyCode == CONTROL) ardrone.landing();  // [Control] 着陸
  } else {
    if      (key == 'a') ardrone.spinLeft();  // [a] 左旋回
    else if (key == 'd') ardrone.spinRight(); // [d] 右旋回
    else if (key == 'w') ardrone.up();        // [w] 上昇
    else if (key == 's') ardrone.down();      // [s] 下降
    else if (key == ' ') {
      if ( centerId >= 0 && targets[centerId].isAlive() ) {
        targets[centerId].shootDown();  // [Space]で画面中心のターゲットを撃破
        shootDownCount++;               // 撃破数をカウント
      }
    }
  }  
}

// キーを離したときに静止命令を送る
void keyReleased() {
  ardrone.stop();  // 静止
}

// 画面上のマーカの中心座標を取得する
PVector getCenterOfMarker(int id) {
  PVector[] vec = ar.getMarkerVertex2D(id);
  return new PVector((vec[0].x + vec[2].x) / 2, (vec[0].y + vec[2].y) / 2);
}

// 指定した座標値が画面の中心付近にあるかどうかを調べる
boolean isCenter(PVector pos) {
  int margin = 20;
  return (pos.x > IMAGE_WIDTH/2  - margin && pos.x < IMAGE_WIDTH/2  + margin && 
          pos.y > IMAGE_HEIGHT/2 - margin && pos.y < IMAGE_HEIGHT/2 + margin);
}

// ロックオン状態を示すリングの描画
void drawLockOnRing() {
  step += 0.1;
  pushStyle();
  stroke(255, 0, 0);
  strokeWeight(5);
  noFill();
  for (int i = 0; i < 4; i++)
    arc(0, 0, 250, 250, step + i * radians(90), step  + i * radians(90) + radians(45) );
  popStyle();
}

// ターゲットの情報を格納したクラス
class Target {
  int     index;
  boolean alive = true;  // 生存フラグ

  Target(int index_) {
    index = index_;
  }

  // ターゲットの撃破
  void shootDown() {
    alive = false;
  }

  // ターゲットが生きているかどうか
  boolean isAlive() {
    return alive;
  }
  
  // ターゲットの描画
  void draw() {
    if (alive) {  // 生きているときは立方体を描画
      fill(0, 0, 255, 80);
      translate(0, 0, 20);
      box(150);
    }
    else {  // 撃破されているときは穴を描画
      imageMode(CENTER);
      image(hole, 0, 0, 2.5 * MARKER_SIZE, 2.5 * MARKER_SIZE);
    }
  }  
}
